/*----------------------------------------------------------------------
  File    : hubness.c
  Contents: computations for hubness phenomenon
  Author  : Christian Borgelt
  History : 2012.01.09 file created
            2012.01.10 hypersphere and hyperball distributions added
            2012.01.11 computation of some hubness indices added
            2012.01.12 generation of grid of holes in hypercubes added
            2012.01.17 generation of grid generalized beyond holes
            2012.01.26 rounding added for quantile computation
            2012.02.06 bug in skewness computation fixed (2/3 -> 3/2)
            2013.08.09 adapted to higher compiler warning level
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <assert.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include "random.h"

#ifndef INFINITY
#define INFINITY   (DBL_MAX+DBL_MAX)
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "hubness"
#define DESCRIPTION "computations for hubness phenomenon"
#define VERSION     "version 1.7 (2014.10.24)         " \
                    "(c) 2012-2014   Christian Borgelt"

/* --- error codes --- */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* file open  failed */
#define E_FREAD      (-3)       /* file read  failed */
#define E_FWRITE     (-4)       /* file write failed (unused) */
#define E_OPTION     (-5)       /* unknown option */
#define E_OPTARG     (-6)       /* missing option argument */
#define E_ARGCNT     (-7)       /* wrong number of arguments */
#define E_MODE       (-8)       /* invalid sampling mode/region */
#define E_CNT        (-9)       /* invalid number of points */
#define E_DIM       (-10)       /* invalid dimensionality */
#define E_NBORS     (-11)       /* invalid number of neighbors */
#define E_UNKNOWN   (-12)       /* unknown error */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- neighbor information --- */
  int      idx;                 /* index    of the data point */
  double   dist;                /* distance to the data point */
} NEIGHBOR;                     /* (neighbor information) */

typedef struct {                /* --- data point --- */
  double   *vec;                /* data vector (coordinates) */
  NEIGHBOR *nbs;                /* list of nearest neighbors */
  int      frq;                 /* freq. in nearest neighbor lists */
} DATAPT;                       /* (data point) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_OPTION   -5 */  "unknown option -%c",
  /* E_OPTARG   -6 */  "missing option argument",
  /* E_ARGCNT   -7 */  "wrong number of arguments",
  /* E_MODE     -8 */  "invalid sampling mode %c'",
  /* E_CNT      -9 */  "invalid number of points %d",
  /* E_DIM     -10 */  "invalid number of dimensions %d",
  /* E_NBORS   -11 */  "invalid number of neighbors %d",
  /* E_UNKNOWN -12 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static const char *prgname;     /* program name for error messages */

DATAPT   *data = NULL;          /* data points + add. information */
double   *vecs = NULL;          /* memory for data vectors */
NEIGHBOR *nnbs = NULL;          /* memory for nearest neighbor lists */
int      *hist = NULL;          /* histogram of neighbor occs. */
FILE     *out  = NULL;          /* output file for hubness data */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

static void delete (void)
{                              /* -- delete data structure */
  if (nnbs) { free(nnbs); } nnbs = NULL;
  if (vecs) { free(vecs); } vecs = NULL;
  if (data) { free(data); } data = NULL;
  hist = NULL;                 /* free all allocated memory */
}  /* delete() */

/*--------------------------------------------------------------------*/

static int create (int cnt, int dim, int knn)
{                              /* --- create data structure */
  int      i;                  /* loop variable */
  double   *p;                 /* to traverse the data points */
  NEIGHBOR *b;                 /* to traverse the neighbors */

  assert((cnt > 0)             /* check the function arguments */
  &&     (dim > 0) && (knn > 0) && (knn < cnt));
  delete();                    /* delete existing data structure */
  data = (DATAPT*)malloc((size_t)cnt *sizeof(DATAPT)
                        +(size_t)cnt *sizeof(int));
  if (!data) return -1;        /* create the data point array */
  hist = (int*)(data +cnt);    /* and the neighor occs. histogram */
  vecs = p = (double*)  malloc((size_t)(cnt *dim) *sizeof(double));
  nnbs = b = (NEIGHBOR*)malloc((size_t)(cnt *knn) *sizeof(NEIGHBOR));
  if (!vecs || !nnbs) { delete(); return -1; }
  for (i = 0; i < cnt; i++) {  /* traverse the data points */
    data[i].vec = p; p += dim; /* set the next data point and */
    data[i].nbs = b; b += knn; /* the next nearest neighbor list */
  }                            /* (organize the memory) */
  return 0;                    /* return 'ok' */
}  /* create() */

/*--------------------------------------------------------------------*/

static void gendata (int mode, int cnt, int dim, int knn, double grid)
{                              /* --- generate data points */
  int    i, k, x;              /* loop variables, flag */
  double c, t;                 /* data point distance, buffer */
  double s, m;                 /* sum/maximum of coordinate values */
  double fac, off;             /* divisor and offset for grid */
  double *p;                   /* to traverse the data points */

  assert((cnt > 0)             /* check the function arguments */
  &&     (dim > 0) && (knn > 0) && (knn < cnt));
  assert(data                  /* check existing data structure */
  &&     data[0].vec && data[0].nbs);

  /* --- generate data points --- */
  if ((mode == 'n')            /* if multivariate normal */
  ||  (mode == 's')            /* or uniform on hypersphere */
  ||  (mode == 'b')            /* or uniform on hyperball */
  ||  (mode == 'x')) {         /* or expanded to hypercube */
    c = 1;                     /* default: unit vector length */
    for (i = 0; i < cnt; i++) {/* traverse the data points */
      p = data[i].vec;         /* get the next data point and */
      for (k = 0; k < dim; k++)/* initialize it with random values */
        p[k] = randn(xrand);   /* that are normally distributed */
      if (mode != 'n') {       /* if hyperball/sphere distribution */
        for (s = k = 0; k < dim; k++)
          s += p[k] *p[k];     /* compute the vector length */
        s = sqrt(s);           /* compute the scaling factor */
        s = (s > 0) ? 1/s : 1; /* for the hypersphere (surface) */
        if (mode != 's')       /* if uniformly dist. on hyperball */
          s *= c = pow(xrand(), 1.0/dim);
        for (k = 0; k < dim; k++)
          p[k] *= s;           /* scale the created random vector */
      }                        /* into hyperball/onto hypersphere */
      if (mode == 'x') {       /* if to expand to cube */
        for (m = k = 0; k < dim; k++)
          if ((t = fabs(p[k])) > m) m = t;
        s = c/m;               /* find the largest coordinate */
        for (k = 0; k < dim; k++)
          p[k] *= s;           /* scale the created random vector */
      }                        /* into a hypercube (expand) */
    } }
  else if (mode == 'j') {      /* hypercube jolted to hyperball */
    for (i = 0; i < cnt; i++){ /* traverse the data points */
      p = data[i].vec;         /* get the next data point */
      for (c = m = k = 0; k < dim; k++) {
        p[k] = 2 *xrand() -1;  /* init. with random values in [-1,1) */
        if ((t = fabs(p[k])) > m) m = t;
        c += t *t;             /* find the largest coordinate */
      }                        /* and compute the vector length */
      c = sqrt(c);             /* compute the scaling factor */
      s = (c > 0) ? m/c : m;   /* for the hyperball */
      for (k = 0; k < dim; k++)
        p[k] *= s;             /* scale the created random vector */
      } }                      /* into a hyperball (jolt) */
  else if (grid > +1) {        /* if grid of (hyper-)cubes */
    m   = floor(grid);         /* get the number of (hyper-)cubes */
    off = grid -m;             /* compute the (hyper-)cube offset */
    fac = 2.0/(grid -1);       /* and the scaling factor */
    for (i = 0; i < cnt; i++){ /* traverse the data points */
      p = data[i].vec;         /* get the next data point and */
      for (k = 0; k < dim; k++) {
        t = xrand() *m;        /* get a random value uniformly */
        c = floor(t);          /* distributed in [0, #cubes]) */
        p[k] = ((t-c) *off +c) *fac -1;
      }                        /* scale the resulting value */
    } }                        /* to the interval [-1,1) */
  else if (grid < -1) {        /* if grid of holes (hypercubic) */
    off = -grid -floor(-grid); /* compute the hole offset */
    fac = 0.5 *(-grid -1);     /* and the scaling factor */
    for (i = 0; i < cnt; i++){ /* traverse the data points */
      p = data[i].vec;         /* get the next data point and */
      for (x = k = 0; k < dim; k++) {
        p[k] = 2 *xrand() -1;  /* init. with random values in [-1,1) */
        t = (p[k] +1) *fac;    /* compute position in hole grid */
        if (t -floor(t) > off) x = -1;
      }                        /* if on a bar, abort the check */
      if (x == 0) { i--; continue; }
    } }                        /* reject data points in holes */
  else {                       /* if uniform on hypercube */
    for (i = 0; i < cnt; i++){ /* traverse the data points */
      p = data[i].vec;         /* get the next data point and */
      for (k = 0; k < dim; k++)/* initialize it with random values */
        p[k] = 2 *xrand() -1;  /* that are uniformly distributed */
    }
  }                            /* (data points are in data[i].vec) */
}  /* gendata() */

/*----------------------------------------------------------------------
Reference for methods to generate points that are uniformly distributed
on a hypersphere or a hyperball:
  R.Y. Rubinstein and D.P. Kroese.
  Simulation and the Monte Carlo Method (2nd edition)
  J. Wiley & Sons, Chichester, United Kingdom 2007
  page 69
----------------------------------------------------------------------*/

static void clear (int cnt, int knn)
{                              /* --- clear nearest neighbor lists */
  int      i, k;               /* loop variables, flag */
  NEIGHBOR *b;                 /* to traverse the neighbors */

  assert((cnt > 0) && (knn > 0)); /* check the function arguments */
  assert(data                  /* check existing data structure */
  &&     data[0].vec && data[0].nbs);

  data[0].frq = 0;             /* clear first frequency counter */
  b = data[0].nbs;             /* get the first nearest neighbor list */
  for (k = 0; k < knn; k++) {  /* and set all distances to infinity */
    b[k].idx = 0; b[k].dist = INFINITY; }
  for (i = 0; ++i < cnt; ) {   /* traverse the other data points */
    data[i].frq = 0;           /* clear the frequency counter */
    memcpy(data[i].nbs, b, (size_t)knn *sizeof(NEIGHBOR));
  }                            /* copy the cleared neighbor list */
}  /* clear() */

/*--------------------------------------------------------------------*/

static void addto (NEIGHBOR *nbs, int n, int idx, double dist)
{                               /* --- add to nearest neighbor list */
  assert(nbs                    /* check the function arguments */
  &&    (n > 0) && (idx >= 0) && (dist >= 0));
  if (dist >= nbs[--n].dist)    /* if farther away than least */
    return;                     /* nearest neighbor in list, abort */
  while ((--n >= 0) && (dist < nbs[n].dist))
    nbs[n+1] = nbs[n];          /* shift all farther neighbors */
  nbs[++n].idx  = idx;          /* store the data point index */
  nbs[  n].dist = dist;         /* and   the data point distance */
}  /* addto() */

/*--------------------------------------------------------------------*/

static int hubness (int mode, int cnt, int dim, int knn,
                    double grid, int mout, FILE *out)
{                              /* --- compute hubness data */
  int      i, k, l, o;         /* loop variables, counters */
  double   c, d, e, t, u;      /* data point distances, buffers */
  double   s, m, x;            /* sum/maximum of coordinate values */
  double   *p, *q;             /* to traverse the data points */
  NEIGHBOR *b;                 /* to traverse the neighbors */

  assert(out                   /* check the function arguments */
  &&    (cnt > 0) && (dim > 0) && (knn > 0) && (knn < cnt));
  if (create(cnt, dim, knn) != 0)
    return -1;                 /* create the data structure */
  gendata(mode, cnt, dim, knn, grid);
  clear(cnt, knn);             /* generate data points */

  /* --- print data points --- */
  if (mout == 'd') {           /* if output mode is data */
    for (k = 0; k < dim; k++){ /* traverse the dimensions */
      if (k > 0) fputc(' ', out);
      fprintf(out, "x%d", k+1);/* print dimension/variable name */
    }                          /* (table header) */
    fputc('\n', out);          /* terminate the output line */
    for (i = 0; i < cnt; i++){ /* traverse the data points */
      p = data[i].vec;         /* get the next data point */
      for (k = 0; k < dim; k++) {
        if (k > 0) fputc(' ', out);
        fprintf(out, "% 14.12f", p[k]);
      }                        /* print the coordinate values */
      fputc('\n', out);        /* terminate the output line */
    }                          /* (print a data table) */
    return 0;                  /* abort the function */
  }                            /* (do not compute hubness data) */

  /* --- compute distances --- */
  for (i = 0; i < cnt; i++) {  /* traverse the data points */
    p = data[i].vec;           /* get the current data point */
    for (k = i; --k >= 0; ) {  /* traverse the preceding data points */
      q = data[k].vec;         /* get the data point */
      for (d = l = 0; l < dim; l++) {
        t = p[l] -q[l]; d += t*t; }
      d = sqrt(d);             /* compute the data point distance */
      addto(data[i].nbs, knn, k, d);
      addto(data[k].nbs, knn, i, d);
    }                          /* if they qualify, add the points */
  }                            /* to their mutual nearest neighbors */

  /* --- count neighbor occurrences --- */
  for (i = 0; i < cnt; i++) {  /* traverse the data points */
    b = data[i].nbs;           /* get the nearest neighbor list */
    for (k = 0; k < knn; k++)  /* and traverse the list */
      data[b[k].idx].frq += 1; /* count data point occurrences */
  }                            /* in nearest neighbor lists */
  memset(hist, 0, (size_t)cnt *sizeof(int));
  for (i = 0; i < cnt; i++)    /* create histogram of numbers of */
    hist[data[i].frq] += 1;    /* occurrences in nearest nb. lists */

  /* --- compute hubness indices --- */
  m = cnt *knn;                /* sum of counters in histogram */
  for (c = d = i = 0; (m > 0) && (i < cnt); i++) {
    m -= hist[i]; t = i-knn; c += e = hist[i] *t *t; d += e *t; }
  c = pow(c/cnt, 1.5);         /* compute second and third moment */
  e = d /(cnt *c);             /* about the mean and then the skew */
  if (mout == 'i') fprintf(out, "%.12g", e);
  else             fprintf(out, "# skewness   : %13.10f\n", e);
  m = cnt;                     /* sum of counters in histogram */
  c = floor(0.001 *cnt +0.5);  /* compute the 0.1%/99.9% quantile */
  d = floor(0.005 *cnt +0.5);  /* and     the 0.5%/99.5% quantile */
  e = floor(0.010 *cnt +0.5);  /* and     the 1.0%/99.0% quantile */
  for (s = t = u = o = k = l = i = 0; i < cnt; i++) {
    x = m -hist[i];            /* traverse the data points */
    if      (m <= c) { s += i *hist[i]; k += hist[i];    }
    else if (x <= c) { s += i *(c-x);   k += (int)(c-x); }
    if      (m <= d) { t += i *hist[i]; l += hist[i];    }
    else if (x <= d) { t += i *(d-x);   l += (int)(d-x); }
    if      (m <= e) { u += i *hist[i]; o += hist[i];    }
    else if (x <= e) { u += i *(e-x);   o += (int)(e-x); }
    if ((m = x) <= 0) break;   /* sum occ. counters in quantiles */
  }                            /* and check for remaining points */
  m  = i/(double)knn;          /* compute the occurrence factors */
  s /= knn *((k > 0) ? k : 1); /* relative to the expected value */
  t /= knn *((l > 0) ? l : 1); /* (i.e. the number of neighbors) */
  u /= knn *((o > 0) ? o : 1); /* and print them */
  if (mout == 'i')             /* if to print only indices */
    fprintf(out, " %.12g %.12g %.12g %.12g", m, s, t, u);
  else {                       /* if to print data point info. */
    fprintf(out, "# hubness 1.0: %13.10fx (maximum)\n",   m);
    fprintf(out, "# hubness 1.1: %13.10fx (top 0.1%%)\n", s);
    fprintf(out, "# hubness 1.2: %13.10fx (top 0.5%%)\n", t);
    fprintf(out, "# hubness 1.3: %13.10fx (top 1.0%%)\n", u);
  }                            /* list indices in comments */
  m = cnt;                     /* sum of counters in histogram */
  k = (2*knn < cnt) ? 2*knn : cnt;
  for (i = 0; i < k; i++)      /* compute the percentage */
    m -= hist[i];              /* of points with at least */
  c = 100 *(m /cnt);           /* twice the number of neighbors */
  k = (3*knn < cnt) ? 3*knn : cnt;
  for (     ; i < k; i++)      /* compute the percentage */
    m -= hist[i];              /* of points with at least */
  d = 100 *(m /cnt);           /* thrice the number of neighbors */
  k = (4*knn < cnt) ? 4*knn : cnt;
  for (     ; i < k; i++)      /* compute the percentage */
    m -= hist[i];              /* of points with at least */
  e = 100 *(m /cnt);           /* thrice the number of neighbors */
  if (mout == 'i')             /* if to print only indices */
    fprintf(out, " %.12g %.12g %.12g", e, d, c);
  else {                       /* if to print data point info. */
    fprintf(out, "# hubness 2.1: %13.10f%% (>= 4k = %d)\n", e, 4*knn);
    fprintf(out, "# hubness 2.2: %13.10f%% (>= 3k = %d)\n", d, 3*knn);
    fprintf(out, "# hubness 2.3: %13.10f%% (>= 2k = %d)\n", c, 2*knn);
  }                            /* list indices in comments */
  if (mout == 'i') {           /* if to print only indices, abort */
    fputc('\n', out); return 0; }

  /* --- print data point data --- */
  fprintf(out, "nbfrq");       /* print a header line */
  if (mout != 's') fprintf(out, "  nearest       length        "
                                "corner        face          stddev");
  if ((mout == 'x') || (mout == 's')) {
    if (mout != 's') fprintf(out, "      ");
    for (k = 0; k < dim; k++)  /* traverse the dimensions */
      fprintf(out, " x%d", k+1);
  }                            /* print dimension/variable name */
  fputc('\n', out);            /* terminate the header line */
  for (i = 0; i < cnt; i++) {  /* traverse the data points */
    p = data[i].vec;           /* get the data point location */
    for (c = d = m = s = l = 0; l < dim; l++) {
      s += t = p[l]; d += t*t; /* traverse the dimensions */
      t = fabs(t);   if (t > m) m = t;
      t = 1-t;       c += t*t; /* compute distance from center and */
    }                          /* from nearest corner and face */
    t = sqrt((d -s*s/dim) /((dim > 1) ? dim : 1));
    x = data[i].nbs[0].dist;   /* compute standard deviation */
    fprintf(out, "%5d", data[i].frq);     /* occs. in neighbor lists */
    if (mout != 's') {         /* if extended data output */
      fprintf(out, " % 13.10f", x);       /* dist. to nearest n.bor */
      fprintf(out, " % 13.10f", sqrt(d)); /* dist. to origin */
      fprintf(out, " % 13.10f", sqrt(c)); /* dist. to nearest corner */
      fprintf(out, " % 13.10f", 1-m);     /* dist. to nearest face */
      fprintf(out, " % 13.10f", t);       /* standard dev. of coords. */
    }
    if (mout == 'x') {         /* if to print coordinates */
      for (k = 0; k < dim; k++)/* traverse the dimensions */
       fprintf(out, " % 14.12f", p[k]);
    }                          /* print the coordinate values */
    fputc('\n', out);          /* terminate the output line */
  }
  return 0;                    /* return 'ok' */
}  /* hubness() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print an error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */

  assert(prgname);              /* check the program name */
  if (code < E_UNKNOWN) code = E_UNKNOWN;
  if (code < 0) {               /* if to report an error, */
    msg = errmsgs[-code];       /* get the error message */
    if (!msg) msg = errmsgs[-E_UNKNOWN];
    fprintf(stderr, "\n%s: ", prgname);
    va_start(args, code);       /* get variable arguments */
    vfprintf(stderr, msg, args);/* print error message */
    va_end(args);               /* end argument evaluation */
    fputc('\n', stderr);        /* terminate the output line */
  }
  #ifndef NDEBUG                /* if compiled as debug version */
  delete();                     /* delete the data structure */
  #endif                        /* (if it has been created) */
  exit(code);                   /* abort the program */
}  /* error() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables */
  char    *s;                   /* to traverse the options */
  char    **optarg = NULL;      /* option argument */
  char    *fname   = NULL;      /* name of output file */
  int     mode     = 'c';       /* sampling mode/region */
  int     cnt      = 1000;      /* number of data points */
  int     dim      = 2;         /* dimensionality of data space */
  int     knn      = 5;         /* number of nearest neighbors */
  double  grid     = 0;         /* grid parameter */
  int     mout     = 'i';       /* output mode for hubness data */
  long    seed;                 /* seed value for random numbers */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] file\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-s#    sampling mode/region        (default: %c)\n", mode);
    printf("       c: uniformly distributed on hypercube [-1,1]^d\n");
    printf("       b: uniformly distributed on hyperball   (r = 1)\n");
    printf("       s: uniformly distributed on hypersphere (r = 1)\n");
    printf("       n: standard normally distributed\n");
    printf("       j: hypercube jolted   into a hyperball\n");
    printf("       x: hyperball expanded into a hypercube\n");
    printf("-n#    number of data points       (default: %d)\n", cnt);
    printf("-d#    number of dimensions        (default: %d)\n", dim);
    printf("-k#    number of nearest neighbors (default: %d)\n", knn);
    printf("-g#    grid parameter for cubes    (default: %g)\n", grid);
    printf("       (positive: grid of hypercubes, negative: holes)\n");
    printf("-S#    seed for random numbers     (default: time)\n");
    printf("-o#    data to print to output     (default: %c)\n", mout);
    printf("       d: only data points (no hubness computation)\n");
    printf("       i: only global hubness measures/indices\n");
    printf("       p: global and pointwise hubness measures\n");
    printf("       x: all hubness and data point information\n");
    printf("       s: k-occurrences and data point coordinates\n");
    printf("file   output file for results\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  seed = (long)time(NULL);      /* get the default seed value */
  for (i = k = 0; ++i < argc; ){/* traverse the arguments */
    s = argv[i];                /* get the next argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate switches */
          case 's': mode = *s++;                  break;
          case 'n': cnt  = (int)strtol(s, &s, 0); break;
          case 'd': dim  = (int)strtol(s, &s, 0); break;
          case 'k': knn  = (int)strtol(s, &s, 0); break;
          case 'g': grid =      strtod(s, &s);    break;
          case 'S': seed =      strtol(s, &s, 0); break;
          case 'o': mout = *s++;                  break;
          default : error(E_OPTION, *--s);        break;
        }                       /* set option variables */
        if (optarg && *s) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-options */
        case  0: fname = s;       break;
        default: error(E_ARGCNT); break;
      }                         /* note filename */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (k != 1) error(E_ARGCNT);  /* and the number of arguments */
  if (strchr("cbsnjx", mode) == NULL) error(E_MODE,  mode);
  if (cnt  <= 0)                      error(E_CNT,   cnt);
  if (dim  <= 0)                      error(E_DIM,   dim);
  if ((knn <= 0) || (knn >= cnt))     error(E_NBORS, knn);
  if (strcmp(fname, "-") == 0) fname = "";
  fprintf(stderr, "\n");        /* terminate the startup message */

  /* --- compute hubness data --- */
  t = clock();                  /* start timer, open output file */
  if (fname && *fname) { out = fopen(fname, "w"); }
  else                 { out = stdout; fname = "<stdout>"; }
  if (!out) error(E_FOPEN, fname);
  fprintf(stderr, "computing hubness data ... ");
  rseed((unsigned)seed);        /* init. random number generator */
  if (hubness(mode, cnt, dim, knn, grid, mout, out) != 0)
    error(E_NOMEM);             /* compute hubness data */
  if (((out != stdout) ? fclose(out) : fflush(out)) != 0)
    error(E_FWRITE, fname);     /* close the output file */
  out = NULL;                   /* clear the file variable */
  fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  #ifndef NDEBUG                /* if compiled as debug version */
  delete();                     /* delete the data structure */
  #endif
  return 0;                     /* terminate the program */
}  /* main() */
